
IsAuto := function(L, mat)
    return IsAutomorphismOfTable(LieTableByBasis(L, Basis(L)), mat);
end;

IsNilpotentDer := function(d)
    return IsNilpotentMat(d![1]);
end;

NilDegree := function(d)
    local s, i;
    if not IsNilpotentDer(d) then return fail; fi;
    s := d![1];
    i := 1;
    while true do
        if s^i = 0*s then return i; fi;
        i := i+1;
    od;
end;

NilDegreeMat := function(m)
    local i;
    for i in [1..Length(m)] do
        if (m-m^0)^i = 0*m then return i; fi;
    od;
    return fail;
end;

IsSandwichDer := function(L, d)
    local s, p, n, t, b, i, j, k, h;

    # get mat, prime and dim
    s := d![1];
    p := Characteristic(LeftActingDomain(L));
    n := Dimension(L);

    # get powers
    t := List([1..p], x -> s^x);

    # check p-nilpotence
    if t[p] <> 0*s then return false; fi;

    # precompute bases
    b := List(t, x -> List(x, y -> y*Basis(L)));

    # check intermediate products
    for i in [1..p-1] do
        if 2*i > p-1 then 
            for k in [1..n] do
                for h in [k+1..n] do
                    if b[i][k]*b[i][h] <> Zero(L) then return false; fi;
                od;
            od;
        fi;
        for j in [i+1..p-1] do
            if i+j > p-1 then 
                for k in [1..n] do
                    for h in [1..n] do
                        if b[i][k]*b[j][h] <> Zero(L) then return false; fi;
                    od;
                od;
            fi;
        od;
    od;
    return true;
end;

AutoBySandwich := function(F, d)
    local p, s, a, k, i;
    p := Characteristic(F);
    s := d![1];
    a := s^0;
    k := 1;
    for i in [1..p-1] do
        k := k*i;
        a := a + (1/k)*s^i;
    od;
    return a;
end;

NilElementsInDer := function(D)
    local p, d, m, x, l, v, s;
    p := Characteristic(LeftActingDomain(D));
    d := Basis(D);
    m := List(d, x -> x![1]);
    x := List([1..Length(d)], 
             i -> Indeterminate(GF(p),Concatenation("x",String(i))));
    l := x*m;
    v := Set(Flat(l^p));
    v := Filtered(v, x -> x <> 0*x);
    v := List(v, ExtExtRep);
    s := SolveSystem(v);
    return List(s, x -> x*d);
end;

ExpAutGroup := function(L)
    local D, n, s, a, i, T;
    D := Derivations(Basis(L));
    Print(Dimension(D), " is dimension of Der \n");
    n := NilElementsInDer(D);
    Print(Length(n), " nilpotent elements \n");
    s := Filtered(n, x -> IsSandwichDer(L,x));
    Print(Length(s), " sandwiches \n");
    a := List(s, x -> AutoBySandwich(LeftActingDomain(L), x));
    return Group(a);
end;

    


