
FieldAuto := function(K, F)
    local R, C, g;
    R := AsField(F,K); 
    C := Basis(R);
    g := GeneratorsOfGroup(GaloisGroup(R))[1];
    return List(C, x -> Coefficients(C, Image(g,x)));
end;

LieAutosTensor := function(L, K)
    local F, g, A, a, b, G;

    # field autos
    F := LeftActingDomain(L);
    g := FieldAuto(K,F); 

    # lie autos
    A := AutomorphismGroup(L);
    a := GeneratorsOfGroup(A); 

    # kronecker
    a := List(a, x -> KroneckerProduct(x,g^0));
    b := [KroneckerProduct(One(A), g)];

    G := Group( Concatenation(a, b));
    SetSize(G, Size(A)*Order(g));
    return G;
end;

AutoByAnnMat := function(F, s)
    local p, a, k, i;
    p := Characteristic(F);
    a := s^0;
    k := 1;
    for i in [1..p-1] do
        k := k*i;
        a := a + (1/k)*s^i;
    od;
    return a;
end;

ExpAutosTensor := function(L, K)
    local F, R, C, g, D, s, a, b;

    # field
    F := LeftActingDomain(L);
    R := AsField(F,K); 
    C := Basis(R);
    g := List(Elements(R), x -> List(C, y -> Coefficients(C, y*x)));

    # ann
    D := Derivations(Basis(L));
    s := Filtered(D, x -> IsSandwichDer(L,x));
    s := List(s, x -> x![1]);

    # kronecker
    a := Concatenation(List(s, x -> List(g, y -> KroneckerProduct(x,y))));
    b := List(a, x -> AutoByAnnMat(F, x));
    return Group(b);
end;

