
l := [[3,1], [7,1],[7,2], [8,1], [8,2], 
      [10,1],
      [14,3], [14,4], [14,5], [14,6],
      [15,2], [15,3], [15,4], [15,5], [15,6], [15,7], [15,8],
      [16,2], [16,3], [16,4], [16,5], [16,6],
      [21,1], [21,2], [21,3], [21,4]];
L := List(l, x -> LieAlgByLibrary(2,x[1],x[2]));

#for i in [1..Length(L)] do
#    s[i] := Set( List( SolvableElements(L[i]), x -> RankElm(L[i],x)));
#od;
#s := List(s, x -> Filtered(x, y -> y <> 0));

s := 
[ [ 2 ], [ 4 ], [ 4 ], [ 4 ], [ 4 ], 
  [ 4, 6 ], # K3(5)
  [ 8 ], [ 8 ], [ 8 ], [ 8 ], [ 8 ], [ 8 ], [ 8 ], [ 8 ], [ 8 ], 
  [ 8 ], [ 8 ], [ 8 ], [ 8 ], [ 8 ], [ 8 ], [ 8 ], 
  [ 6, 10, 12 ], # K3(7)
  [ 12 ], [ 14 ], [ 12 ] ];

RankElm := function( L, l )
    return RankMat(AdjointMatrix(Basis(L),l));
end;

SolvableRad := function(L)
    return Subalgebra( L, SolvableElements(L) );
end;
   
IsCano := function( mat )
    local l, i, j, t;

    # special cases
    if mat = 0*mat then return true; fi;
    if mat = mat^0 then return true; fi;

    # check off-diagonal
    l := Length(mat);
    for i in [1..l] do
        for j in [1..l] do
            if i <> j and mat[i][j] <> 0*mat[i][j] then return false; fi;
        od;
    od;

    # check diagonal
    t := First([1..l], i -> mat[i][i] = 0*mat[i][i]);
    return ForAll([t..l], i -> mat[i][i] = 0*mat[i][i]);
end;

TryOrbits := function( n, l )
    local G, H, D, M, act, o, i, j, w, s;
    G := GL(l, 2);
    H := GL(n-l, 2);
    D := DirectProduct(G,H);
    M := MatrixSpace( GF(2), 2*l, n-l );
    act := function( mat, tup )
        local a, b, m, w;
        m := Image(Projection(D,1), tup);
        w := Image(Projection(D,2), tup); 
        a := mat{[1..l]};
        a := m^-1 * a * w;
        b := mat{[l+1..2*l]};
        b := TransposedMat(m) * b * TransposedMat(w^-1);
        return Concatenation(a,b);
    end;
    o := Orbits(D, Elements(M), act);
    for i in [1..Length(o)] do
        w := Filtered(o[i], x -> IsCano(x{[1..l]}));
        j := First([1..Length(w)], x -> IsCano(w[x]{[l+1..2*l]}));
        if IsInt(j) then 
            Print("good case \n");
            Display(w[j]);
            Print("\n");
        else
            Print("bad case \n");
            for s in w do
                Display(s); 
                Print("\n");
            od;
            Error();
        fi;
    od;
end;

    
GradingByElm := function(L)
    local a, w, b, v, B, A, C, M, U, W, u, h, d, l;

    l := SolvableElements(L)[2];
    d := Dimension(L);
    a := TransposedMat(AdjointMatrix(Basis(L), l));
    w := NullspaceMat(a-a^0);
    v := NullspaceMat(a);
    b := Concatenation(w,v);
    l := Length(w);
 
    B := List(b, x -> LinearCombination(Basis(L), x));
    A := List(B, x -> TransposedMat(AdjointMatrix(Basis(L), x)));
    C := List(A, x -> b * x * b^-1);

    M := MatrixSpace(GF(2), d, d);
    U := Subspace(M, C{[1..l]});
    W := Subspace(M, C{[l+1..d]});

    u := Set(List(Elements(U), x -> MinimalPolynomial(GF(2),x)));
    h := Set(List(Elements(W), x -> 
           [MinimalPolynomial(GF(2),x{[1..l]}{[1..l]}),
            MinimalPolynomial(GF(2),x{[l+1..d]}{[l+1..d]})]));
    return rec(odd := u, even := h);
end;

