
DisplayPT := function(T)
    local i, j, k;
    for i in [1..Length(T)] do
        Print("\n",i,": \n");
        for j in [i+1..Length(T)] do
            Print(" ");
            for k in [1..Length(T[i][j])] do
                if IsFFE(T[i][j][k]) then 
                    Print(IntFFE(T[i][j][k])," ");
                else
                    Print("* ");
                fi;
            od;
            Print("\n");
        od;
    od;
    Print("\n");
end;

SubWeightPT := function(t)
    return Length(Filtered(Flat(t), x -> x = false));
end;

GetEntryPT := function(T,i,j)
    if i=j then 
        return 0*T[1][2];
    elif i>j then 
        return T[j][i];
    else
        return T[i][j];
    fi;
end;
            
LieMatGenPT := function(T)
    local mats, i, mat;
    mats := [];
    for i in [1..Length(T)] do
        mat := List([1..Length(T)], j -> GetEntryPT(T,i,j));
        if SubWeightPT(mat)=0 then Add(mats, mat); fi;
    od;
            
    return LieAlgebra(GF(2), mats);
end;

JacobiCasePT := function(T,i,j,k,h,n)
    local s, p, l, a, b, t;
    s := [];
    p := false;
    for l in [1..n] do
        a := GetEntryPT(T,i,j)[l];
        b := GetEntryPT(T,l,k)[h];
        if IsBool(a) and IsBool(b) then 
            return true;
        elif not IsBool(a) and not IsBool(b) then
            Add(s, a*b);
        elif a = Zero(GF(2)) or b = Zero(GF(2)) then 
            Add(s, Zero(GF(2)));
        elif not IsBool(a) and IsBool(p) then 
            p := [l,k,h,a];
        elif not IsBool(b) and IsBool(p) then 
            p := [i,j,l,b];
        else
            return true;
        fi;
 
        a := GetEntryPT(T,j,k)[l];
        b := GetEntryPT(T,l,i)[h];
        if IsBool(a) and IsBool(b) then 
            return true;
        elif not IsBool(a) and not IsBool(b) then
            Add(s, a*b);
        elif a = Zero(GF(2)) or b = Zero(GF(2)) then 
            Add(s, Zero(GF(2)));
        elif not IsBool(a) and IsBool(p) then 
            p := [l,i,h,a];
        elif not IsBool(b) and IsBool(p) then 
            p := [j,k,l,b];
        else
            return true;
        fi;
         
        a := GetEntryPT(T,k,i)[l];
        b := GetEntryPT(T,l,j)[h];
        if IsBool(a) and IsBool(b) then 
            return true;
        elif not IsBool(a) and not IsBool(b) then
            Add(s, a*b);
        elif a = Zero(GF(2)) or b = Zero(GF(2)) then 
            Add(s, Zero(GF(2)));
        elif not IsBool(a) and IsBool(p) then 
            p := [l,j,h,a];
        elif not IsBool(b) and IsBool(p) then 
            p := [k,i,l,b];
        else
            return true;
        fi;
    od;

    t := Sum(s);
    if IsBool(p) then return t = Zero(GF(2)); fi;
    #Print("  Jacobi: set value \n");
    if p[1]<p[2] then 
        T[p[1]][p[2]][p[3]] := t/p[4];
    elif p[1]>p[2] then
        T[p[2]][p[1]][p[3]] := t/p[4];
    fi;
    return true;
end;

JacobiPT := function(T)
    local i, j, k, n, h, jac;
    n := Length(T);
    for i in [1..n] do
        for j in [i+1..n] do
            for k in [j+1..n] do
                for h in [1..n] do
                    if not JacobiCasePT(T,i,j,k,h,n) then return false; fi;
                od;
            od;
        od;
    od;
    return true;
end;

JacobiPrintCasePT := function(T,i,j,k,h,n)
    local a, b, l, c;
    c := [];
    for l in [1..n] do

        a := GetEntryPT(T,i,j)[l];
        b := GetEntryPT(T,l,k)[h];
        if IsBool(a) and IsBool(b) then 
            Add(c, 2);
        elif a = Zero(GF(2)) and IsBool(b) then 
            Add(c, 0);
        elif b = Zero(GF(2)) and IsBool(a) then 
            Add(c, 0);
        elif not IsBool(a) and IsBool(b) then 
            Add(c, 1);
        elif not IsBool(b) and IsBool(a) then 
            Add(c, 1);
        fi;
 
        a := GetEntryPT(T,j,k)[l];
        b := GetEntryPT(T,l,i)[h];
        if IsBool(a) and IsBool(b) then 
            Add(c, 2);
        elif a = Zero(GF(2)) and IsBool(b) then 
            Add(c, 0);
        elif b = Zero(GF(2)) and IsBool(a) then 
            Add(c, 0);
        elif not IsBool(a) and IsBool(b) then 
            Add(c, 1);
        elif not IsBool(b) and IsBool(a) then 
            Add(c, 1);
        fi;

        a := GetEntryPT(T,k,i)[l];
        b := GetEntryPT(T,l,j)[h];
        if IsBool(a) and IsBool(b) then 
            Add(c, 2);
        elif a = Zero(GF(2)) and IsBool(b) then 
            Add(c, 0);
        elif b = Zero(GF(2)) and IsBool(a) then 
            Add(c, 0);
        elif not IsBool(a) and IsBool(b) then 
            Add(c, 1);
        elif not IsBool(b) and IsBool(a) then 
            Add(c, 1);
        fi;

    od;
    if Length(c) > 0 then 
        Print(i," ",j," ",k," ",h," yields ",Collected(c),"\n");
    fi;
end;

JacobiPrintPT := function(T)
    local i, j, k, n, h, jac;
    n := Length(T);
    for i in [1..n] do
        for j in [i+1..n] do
            for k in [j+1..n] do
                for h in [1..n] do
                    JacobiPrintCasePT(T,i,j,k,h,n);
                od;
            od;
        od;
    od;
end;

GetBlock := function(n,e,i,j,B)
    local ind, l, k;
    ind := [];
    for l in [1..n] do
        for k in [l+1..n] do
            if e[l] = i and e[k] = j then 
                Add(ind, Coeff(n,l,k));
            fi;
        od;
    od;
    return B{ind}{ind};
end;

SetUpPT := function( A )
    local F, n, V, f, g, r, K, d, b, i, j, h, k, e, B, C, new,T,l;

    # get field and dim
    F := GF(2);
    n := Length(A);
    V := F^n;

    # b[i][j] contains the possible blocks for [Vi,Vj] with j>=i
 
    # set up
    f := CharacteristicPolynomial(A);
    g := Collected(Factors(f));
    r := Length(g);
    K := SplittingField(f);
    d := List(g, x -> RootsOfUPol(K,x[1]));
    A := JNFByPolys(Flat(List(g, x -> List([1..x[2]], y-> x[1]))),GF(2));

    # set up
    b := List([1..r], x -> []);

    # V1 is the 0-Eigenraum with [V1,Vi] in Vi
    for i in [1..r] do b[1][i] := [i]; od;

    # for the remaining blocks consider eigenvalues
    for i in [2..r] do
        for j in [i..r] do
            e := Flat(List(d[i], x -> x+d[j]));
            b[i][j] := Filtered([1..r], x -> Length(Intersection(e,d[x]))>0);
        od;
    od;

    # add wedge action
    B := ExtAction(IdentityMat(n,GF(2)),A);
    e := List(g, x -> Degree(x[1])*x[2]);
    l := Flat(List([1..Length(e)], x -> List([1..e[x]], y -> x))); 

    # go through blocks
    for i in [1..r] do
        for j in [i..r] do
            C := GetBlock(n,l,i,j,B);
            if Length(C)>0 then 
                h := MinimalPolynomial(GF(2),C);
                k := Set(Factors(h));
                new := Filtered([1..r], x -> g[x][1] in k);
                b[i][j] := Intersection(b[i][j],new);
            fi;
        od;
    od;
            
    # set up structure constants table
    T := List([1..n], x -> []);

    # add some obvious values
    for i in [2..n] do
        T[1][i] := A[i];
    od;

    # go through
    for i in [2..n] do
        for j in [i+1..n] do
            k := b[l[i]][l[j]];
            T[i][j] := ShallowCopy(Zero(V));
            for h in [1..n] do
                if l[h] in k then 
                    T[i][j][h] := false;
                fi;
            od;
        od;
    od;
                
    # evaluate Jacobi
#    if not JacobiPT(T) then return false; fi;
    return rec( tab := T, bls := l );
end;

RelPosPT := function(T)
    local e, f, n, ind, j, k, h, c;

    # catch indices
    n := Length(T);
    c := 1;

    # determine positions of unkowns
    ind := [];
    for j in [1..n] do
        for k in [j+1..n] do
            for h in [1..n] do
                if T[j][k][h] = false then 
                    Add(ind, c);
                fi;
                c := c+1;
            od;
        od;
    od;

    # that's it
    return ind; 
end;

SubsPT := function(T,v)
    local e, f, n, S, c, j, k, h;

    # catch indices
    n := Length(T);
    S := StructuralCopy(T);

    # add in v
    c := 1;
    for j in [1..n] do
        for k in [j+1..n] do
            for h in [1..n] do
                if T[j][k][h] = false and IsBool(v) then 
                    S[j][k][h] := Zero(GF(2));
                    c := c+1;
                elif T[j][k][h] = false then 
                    if not IsBool(v[c]) then S[j][k][h] := v[c]; fi;
                    c := c+1;
                fi;
            od;
        od;
    od;

    # that's it
    return S;
end;

CheckPT := function(T)
    local i, m;

    # check Jacobi and fill up values
    if not JacobiPT(T) then return false; fi; 

    # check matrices
    for i in [1..Length(T)] do
        m := List([1..Length(T)], x -> GetEntryPT(T,i,x));
        if SubWeightPT(m)=0 then 
            if m=0*m or not IsLieRelevant1(m) then 
                return false;
            fi;
        fi;
    od;

    return true;
end;

CompletePTByBruteForce := function(arg)
    local T, n, m, V, e, f, v, w, S, L, l, s;

    # catch arguments
    T := arg[1];
    n := RelPosPT(T);
    if Length(arg) = 2 then m := arg[2]; else m := [1..Length(n)]; fi;
    V := GF(2)^Length(m);
    e := []; f := [];

    # loop
    for v in V do

        # replace unknowns by values
        w := List([1..Length(n)], x -> false); w{m} := v;
        S := SubsPT(T,w);

        # do some simple checks 
        if CheckPT(S) then 
            L := LieMatGenPT(S);
            l := Dimension(L);
            s := List(S, SubWeightPT);
            Print(" got ",s, " with dim ",l,"\n");
            if l >= Length(T) then 
                L := CheckSubalgebra(false, L);
                if Length(L.new)>0 then 
                    Error("new algebra");
                else
                    Add(e, L.res);
                fi;
            elif s <> 0*s then 
                Add(f,S); 
            fi;
        else
            Print(" failure \n");
        fi;
    od;

    return rec( tabs := f, lies := e );
end;


CompletePTByGLAction := function(T)
    local C, d, n, ind, act, orb, vec, sub, t, e, v, S, L, l, s, dif;

    # get acting group 
    Print("compute action \n");
    C := GLOnTables(Length(T));
    d := Length(One(C));
    n := Length(T);

    # get relevant indices and subspace
    Print("indices \n");
    ind := RelPosPT(T);
    vec := IdentityMat(d,GF(2)){ind};
    S := Stabilizer(C, vec, OnBases);

    # stabilize elm in quotient
    dif := Difference([1..d], ind);
    vec := IdentityMat(d,GF(2)){dif};
   


Error();

    # determine orbits and add zero
    orb := MatOrbs(act, Length(act[1]), GF(2));
    Add(orb, 0*orb[1]);

    # loop over orbs
    t := []; e := [];
    for v in orb do

        # replace unknowns by values
        S := SubsPT(T,v);

        # do some simple checks 
        if CheckPT(S) then 
            L := LieMatGenPT(S);
            l := Dimension(L);
            s := List(S, SubWeightPT);
            Print(" got ",s, " with dim ",l,"\n");
            if l >= Length(T) then 
                L := CheckSubalgebra(false, L);
                if Length(L.new)>0 then 
                    Error("new algebra");
                else
                    Add(e, L.res);
                fi;
            elif s <> 0*s then 
                Add(t[l],S); 
            fi;
        else
            Print(" failure \n");
        fi;
    od;

    return rec( tabs := t, lies := e );
end;


