
TypeOfElement := function(L, g)
    local F, B, b, m;
    F := LeftActingDomain(L);
    B := Basis(L);
    b := BasisVectors(B);
    m := TransposedMat(AdjointMatrix(B,g)); 
    return MinimalPolynomial(m);
end;

PolynomialsForType := function(L, type)
    local n, p, m, x, l, y, v;

    # set up
    n := Dimension(L);
    p := Characteristic(LeftActingDomain(L));
    m := LieAdjointModule(L).generators;
    x := List([1..n], i -> Indeterminate(GF(2),Concatenation("x",String(i))));

    # get polys
    l := x*m;
    v := Set(Flat(Value( type, l )));
    v := Filtered(v, x -> x <> 0*x);
    return v;
end;


ElementsByType := function(L, f)
    local pol, sol;
    pol := PolynomialsForType( L, f);
    sol := SolveSystem( List(pol, ExtExtRep) ); 
    sol := List(sol, x -> x * Basis(L));
    return sol;
end;

NilpotentElements := function(L)
    local y, pol, sol;
    y := Indeterminate(GF(2),"y": old);
    pol := PolynomialsForType( L, y^2);
    sol := SolveSystem(List(pol, ExtExtRep));
    sol := List(sol, x -> x * Basis(L));
    return sol;
end;

SolvableElements := function(L)
    local y, pol, sol;
    y := Indeterminate(GF(2),"y": old);
    pol := PolynomialsForType( L, y^2+y);
    sol := SolveSystem(List(pol, ExtExtRep));
    sol := List(sol, x -> x * Basis(L));
    return sol;
end;

NumberOfNilpotentElements := function(L)
    local y, pol;
    y := Indeterminate(GF(2),"y": old);
    pol := PolynomialsForType( L, y^2);
    return Length(SolveSystem(List(pol, ExtExtRep)));
end;

NumberOfSolvableElements := function(L)
    local y, pol;
    y := Indeterminate(GF(2),"y": old);
    pol := PolynomialsForType( L, y^2-y);
    return Length(SolveSystem(List(pol, ExtExtRep)));
end;

NumberOfElementsByType := function(L, f)
    local pol;
    pol := PolynomialsForType( L, f );
    return Length(SolveSystem(List(pol, ExtExtRep)));
end;

NumberOfElementsWithType := function(L)
    return NumberOfElementsByType( L, L!.type );
end;

