
SamplePols := function(L)
    local pols, stat, i, l, m, f, j;
    pols := [];
    for i in [1..100*Dimension(L)] do
        l := Random(L);
        m := TransposedMat(AdjointMatrix(Basis(L), l));
        f := MinimalPolynomial(m);
        AddSet(pols, f);
    od;
    return pols;
end;

SampleElms := function(L, s)
    local i, l, m, f, j;
    for i in [1..100*Dimension(L)] do
        l := Random(L);
        m := TransposedMat(AdjointMatrix(Basis(L), l));
        f := MinimalPolynomial(m);
        j := Position(s.pols, f);
        if IsInt(j) then Add(s.elms[j], l); fi;
    od;
    return s;
end;

FindGensInSet := function(L, e)
    local gens, i, g, U, h;
    gens := false;
    i := 0;
    while gens = false or i <= 10*Dimension(L) do
        i := i+1;
        g := [];
        U := Subalgebra(L, g);
        repeat
            h := Random(e);
            if not h in U then 
                Add(g, h);
                U := SubalgebraNC(L, g);
            fi;
        until Dimension(U) = Dimension(L);
        if Length(g) = 2 then return g; fi;
        if gens = false then gens := g; fi;
        if IsList(gens) and Length(g) < Length(gens) then gens := g; fi;
    od;
    return gens;
end;


SamplePolsAndElms := function(L)
    local pols, elms, todo, i, l, m, f, j, I, degs;

    # get samples
    pols := [];
    elms := [];
    todo := [];
    for i in [1..100*Dimension(L)] do
        l := Random(L);
        m := TransposedMat(AdjointMatrix(Basis(L), l));
        f := MinimalPolynomial(m);
        j := Position(pols, f);
        if IsBool(j) then 
            Add(pols, f);
            Add(elms, [l]);
            Add(todo, 1);
        elif todo[j] < Dimension(L) then 
            Add(elms[j], l);
            todo[j] := Dimension(Subalgebra(L, elms[j]));
        fi;
    od;
    I := Filtered([1..Length(todo)], x -> todo[x] = Dimension(L));
    pols := pols{I};
    elms := elms{I};

    # sort
    degs := List(pols, Degree);
    I := [1..Length(pols)];
    SortParallel(degs, I);

    # return
    return rec( pols := pols{I}, elms := elms{I} );
end;

AddGensAndType := function(L)
    local samp, gens, i, m;

    # sample
    samp := SamplePolsAndElms(L);
    gens := [];

    # search for short gen-set
    for i in [1..Length(samp.pols)] do
        gens[i] := FindGensInSet(L, samp.elms[i]);
        if Length(gens[i]) = 2 then 
            L!.gens := List(gens[i], x -> Coefficients(Basis(L),x));
            L!.type := samp.pols[i];
            return;
        fi;
    od;
    m := List(gens, Length);
    i := Position(m, Minimum(m));
    L!.gens := List(gens[i], x -> Coefficients(Basis(L),x));
    L!.type := samp.pols[i];
end;

DetermineGensAndType := function(L)
    local samp, d, g, e, i, j;
    samp := SamplePolsAndElms(L);
    d := []; g := []; e := [];
    for i in [1..Length(samp.pols)] do
        g[i] := FindGensInSet(L, samp.elms[i]);
        e[i] := ElementsByType(L, samp.pols[i]);
        d[i] := Length(e[i])^Length(g[i]);
        Print(" found ",d[i],"\n");
    od;
    j := First([1..Length(samp.pols)], x -> d[x] = Minimum(d));
    L!.type := samp.pols[j];
    L!.gens := List(g[j], x -> Coefficients(Basis(L),x));
    L!.lens := [Length(e[j]), Length(g[j])];
end;
    
SpecOrbits := function(G, L, s)
    local O, i, c, j, k, o;
    O := List([1..Length(s.elms)], x -> []);
    for i in [1..Length(s.elms)] do
        c := List(s.elms[i], x -> Coefficients(Basis(L), x));
        for j in [1..Length(c)] do
            k := List(O[i], x -> c[j] in x);
            if ForAll(k, x -> x = false) then
                o := Orbits(G,c[j],OnRight);
                Add(O[i], o);
            fi;
        od;
    od;
    return O;
end;

NewFindGensInSet := function(L, e, f)
    local gens, i, g, U, h;
    gens := false;
    i := 0;
    while gens = false or i <= 10*Dimension(L) do
        i := i+1;
        g := [f];
        U := Subalgebra(L, g);
        repeat
            h := Random(e);
            if not h in U then
                Add(g, h);
                U := SubalgebraNC(L, g);
            fi;
        until Dimension(U) = Dimension(L);
        if Length(g) = 2 then return g; fi;
        if gens = false then gens := g; fi;
        if IsList(gens) and Length(g) < Length(gens) then gens := g; fi;
    od;
    return gens;
end;

NewTry := function(L, s)
    local S, i, g, T;
    S := Subalgebra(L,s);
    if Dimension(S) = Dimension(L) then 
        Print("single case \n");
        return FindGensInSet(L,s);
    else
        for i in [1..100] do
            g := Random(L);
            if not g in S then 
               T := Subalgebra(L, Concatenation(Basis(S),[g]));
               Print(i," yields ",Dimension(T),"\n");
               if Dimension(T) = Dimension(L) then 
                   Print("multi case \n");
                   return NewFindGensInSet(L, s, g);
               fi;
            fi;
        od;
    fi;
end;
 
            
