
MultMono := function( a, b )
    local c, i;
    if a = 0 or b = 0 then return 0; fi;
    c := 1;
    for i in [1..Length(a)] do
        c := c * Binomial(a[i]+b[i],a[i]) mod 2;
        if c = 0 then return 0; fi;
    od;
    return a+b;
end;

ApplyDerivation := function( i, a )
    local b;
    if a[i] = 0 then return 0; fi;
    b := StructuralCopy(a);
    b[i] := b[i]-1;
    return b;
end;

SimpleLieAlgebraTypeBR2 := function( n, F )
    local p, L, l, d, e, f, T, i, j, c, a, b, a1, b1, a2, b2, r, s;

    if Length(n) <> 2 then return fail; fi;

    # get W(2,n)
    L := SimpleLieAlgebraTypeW( n, F );
    l := L[2];
    L := L[1];
    d := Dimension(L);
    e := d/2;

    # get A(2,n)
    f := PolynomialBasis( n, 2 );

    # emtpy table
    T := EmptySCTable( 2*d, Zero(F), "antisymmetric" );
 
    # [W,W] 
    for i in [1..d] do
        for j in [i+1..d] do
            c := Coefficients(Basis(L), Basis(L)[i]*Basis(L)[j]);
            SetEntrySCTable( T, i, j, WordByExponents(c));
        od;
    od;

    # [A,A]
    for i in [d+1..d+e] do 
        for j in [i+1..d+e] do
            a := f[i-d];
            b := f[j-d];
            a1 := ApplyDerivation(1,a);
            a2 := ApplyDerivation(2,a);
            b1 := ApplyDerivation(1,b);
            b2 := ApplyDerivation(2,b);
            r := Position(f, MultMono(a2,b1));
            s := Position(f, MultMono(a1,b2));
            c := List([1..2*d], x -> Zero(F));
            if IsInt(r) then c[d+e+r] := One(F); fi;
            if IsInt(s) then c[d+e+s] := c[d+e+s] + One(F); fi;
            SetEntrySCTable( T, i, j, WordByExponents(c));
        od;
    od;

    # [ol{A},ol{A}] = 0
          
    # [W,A] 
    for i in [1..d] do
        for j in [d+1..d+e] do
            a := l[i];
            b := f[j-d];
            r := ApplyDerivation(a[2], b);
            s := Position(f, MultMono(a[1],r));
            c := List([1..2*d], x -> Zero(F));
            if IsInt(s) then c[d+s] := One(F); fi;
            SetEntrySCTable( T, i, j, WordByExponents(c));
        od;
    od;

    # [W, ol{A}]
    for i in [1..d] do
        for j in [d+e+1..d+2*e] do
            a := l[i];
            b := f[j-d-e];
            r := MultMono(a[1],b);
            c := List([1..2*d], x -> Zero(F));
            if IsInt(Position(f,r)) then 
                s := Position(f,ApplyDerivation(a[2], r));
                if IsInt(s) then c[d+e+s] := One(F); fi;
            fi;
            SetEntrySCTable( T, i, j, WordByExponents(c));
        od;
    od;

    # [A, ol{A}]
    for i in [d+1..d+e] do
        for j in [d+e+1..d+2*e] do
            a := f[i-d];
            b := f[j-d-e];
            a1 := ApplyDerivation(1,a);
            a2 := ApplyDerivation(2,a);
            r := MultMono(b,a1);
            s := MultMono(b,a2);
            r := Position(l, [r,2]);
            s := Position(l, [s,1]);
            c := List([1..2*d], x -> Zero(F));
            if IsInt(r) then c[r] := One(F); fi;
            if IsInt(s) then c[s] := c[s] + One(F); fi;
            SetEntrySCTable( T, i, j, WordByExponents(c));
        od;
    od;

#Print(TestJacobi(T),"\n");
            
    L := LieAlgebraByStructureConstants( F, T );
    L := LieDerivedSubalgebra(L/LieCenter(L));
    SetName(L, Concatenation("BR2",StringList(n)));
    return L;
end;

SimpleLieAlgebraTypeBR1 := function( n, F )
    local L, U, l, f, d, e, T, i, j, k, c, a, b, r, a1,a2,a3,b1,b2,b3, v;

    if Length(n) <> 3 then return fail; fi;

    # get W(3,n) and S(3,n)
    L := SimpleLieAlgebraTypeW( n, F );
    U := SimpleLieAlgebraTypeS( n, F );
    l := L[2];
    L := L[1];

    # get A(3,n)
    f := PolynomialBasis( n, 2 );

    # dimensions
    d := Dimension(U);
    e := Length(f);

    # emtpy table
    T := EmptySCTable( d+2*e, Zero(F), "antisymmetric" );
 
    # [W,W] 
    for i in [1..d] do
        for j in [i+1..d] do
            c := Coefficients(Basis(U), Basis(U)[i]*Basis(U)[j]);
            SetEntrySCTable( T, i, j, WordByExponents(c));
        od;
    od;

    # [A,A] = 0
    # [ol{A},ol{A}] = 0
          
    # [W,A] 
    for i in [1..d] do
        for j in [d+1..d+e] do
            c := List([1..d+2*e], x -> Zero(F));
            b := f[j-d];
            a := Coefficients(Basis(L), Basis(U)[i]);
            for k in [1..Length(a)] do
                if a[k] <> Zero(F) then 
                    r := ApplyDerivation(l[k][2], b);
                    r := MultMono(l[k][1], r);
                    r := Position(f, r);
                    if IsInt(r) then c[d+r] := c[d+r]+One(F); fi;
                fi;
            od;
            SetEntrySCTable( T, i, j, WordByExponents(c));
        od;
    od;

    # [W, ol{A}]
    for i in [1..d] do
        for j in [d+e+1..d+2*e] do
            c := List([1..d+2*e], x -> Zero(F));
            b := f[j-d-e];
            a := Coefficients(Basis(L), Basis(U)[i]);
            for k in [1..Length(a)] do
                if a[k] <> Zero(F) then 
                    r := ApplyDerivation(l[k][2], b);
                    r := MultMono(l[k][1], r);
                    r := Position(f, r);
                    if IsInt(r) then c[d+e+r] := c[d+e+r]+One(F); fi;
                fi;
            od;
            SetEntrySCTable( T, i, j, WordByExponents(c));
        od;
    od;

    # [A, ol{A}]
    for i in [d+1..d+e] do
        for j in [d+e+1..d+2*e] do
            a := f[i-d];
            b := f[j-d-e];
            a1 := ApplyDerivation(1,a);
            a2 := ApplyDerivation(2,a);
            a3 := ApplyDerivation(3,a);
            b1 := ApplyDerivation(1,b);
            b2 := ApplyDerivation(2,b);
            b3 := ApplyDerivation(3,b);
            
            v := Zero(L);
            r := Position(l, [MultMono(a1,b2),3]);
            if IsInt(r) then v := v + Basis(L)[r]; fi;
            r := Position(l, [MultMono(a1,b3),2]);
            if IsInt(r) then v := v + Basis(L)[r]; fi;
            r := Position(l, [MultMono(a2,b1),3]);
            if IsInt(r) then v := v + Basis(L)[r]; fi;
            r := Position(l, [MultMono(a2,b3),1]);
            if IsInt(r) then v := v + Basis(L)[r]; fi;
            r := Position(l, [MultMono(a3,b1),2]);
            if IsInt(r) then v := v + Basis(L)[r]; fi;
            r := Position(l, [MultMono(a3,b2),1]);
            if IsInt(r) then v := v + Basis(L)[r]; fi;

            c := Coefficients(Basis(U), v);
            SetEntrySCTable( T, i, j, WordByExponents(c));
        od;
    od;

#Print(TestJacobi(T),"\n");
            
    L := LieAlgebraByStructureConstants( F, T );
    L := LieDerivedSubalgebra(L/LieCenter(L));
    SetName(L, Concatenation("BR1",StringList(n)));
    return L;
end;

