
LieAlgebraByLibrary := function( p, dim, nr )
    local SLD;
    if p = 2 then SLD := SLDB2; elif p = 3 then SLD := SLDB3; fi;
    if not IsBound(SLD[dim]) or not IsBound(SLD[dim][nr]) then 
        return fail;
    fi;
    return SLD[dim][nr];
end;

#############################################################################
AddNameDB := function( p, dim, nr, name )
    local file;
    file := Concatenation(DBPATH,String(p),"/Dim_",String(dim),"_Nr_",
                          String(nr));
    AppendTo(file, "Add(LIEALG.names, \"");
    AppendTo(file, name);
    AppendTo(file, "\");\n");
end;

AddPresDB := function(p, dim, nr)
    local file, c, L, r, i;;

    # get file
    file := Concatenation(DBPATH,String(p),"/Dim_",String(dim),"_Nr_",
                          String(nr));
   
    # get rels
    L := LieAlgebraByLibrary(p,dim,nr);
    r := PresentationByGens(L, L!.gens);

    # write into file
    AppendTo(file, "LIEALG.rels := [\n");
    for i in [1..Length(r)] do
        AppendTo(file, r[i]);
        AppendTo(file, ",\n");
    od;
    AppendTo(file, "];\n\n");
end;

#############################################################################
CheckInDB := function(L, nr)
    local d, p, f, file, T, c, name, SLT, i, g, r;

    # set up
    d := Dimension(L);
    p := Characteristic(LeftActingDomain(L));
    if p = 2 then SLT := SLTB2; elif p = 3 then SLT := SLTB3; fi;
    f := Concatenation(DBPATH,String(p),"/");

    # increase counter if necessary
    if not IsInt(nr) then 
        SLT[d] := SLT[d]+1;
        PrintTo(Concatenation(f, "sltb.gi"), "SLTB",String(p)," := \n");
        AppendTo(Concatenation(f, "sltb.gi"), SLT);
        AppendTo(Concatenation(f, "sltb.gi"), "; \n");
        nr := SLT[d];
    fi;

    # the filename
    file := Concatenation(f,"Dim_",String(d),"_Nr_",String(nr));

    # init the file
    PrintTo(file,"LIEALG := rec( ); \n\n");

    #######################################################################
    # add names
    AppendTo(file, "LIEALG.names := [ ]; \n");
    if IsBound(L!.names) then 
        for name in L!.names do AddNameDB( p, d, nr, name ); od;
    fi;
    AppendTo(file, "\n");

    #######################################################################
    # the structure constants
    T := LieTableByBasis(L, Basis(L));
    T := List(T, x -> List(x, y -> IntVecFFE(y)));
    AppendTo(file, "LIEALG.table := \n");
    AppendTo(file, T);
    AppendTo(file, "\n");
    AppendTo(file,"*One(GF(",String(p),")); \n\n");

    #######################################################################
    # number of nilpotent elements (ad(x))^2 = 0)
    if IsBound(L!.nrnil) then 
        c := L!.nrnil;
    else
        c := NumberOfNilpotentElements(L);
    fi;
    AppendTo(file, "LIEALG.nrnil := ");
    AppendTo(file, c);
    AppendTo(file, "; \n");
    
    #######################################################################
    # number of solvable elements (ad(x))^2 = -ad(x)
    if IsBound(L!.nrsol) then 
        c := L!.nrsol;
    else
        c := NumberOfSolvableElements(L);
    fi;
    AppendTo(file, "LIEALG.nrsol := ");
    AppendTo(file, c);
    AppendTo(file, "; \n");

    #######################################################################
    # add gens and type
    if not IsBound(L!.gens) then AddGensAndType(L); fi;
    g := L!.gens;
    AppendTo(file,"LIEALG.gens := [\n");
    for i in [1..Length(g)] do
        AppendTo(file,IntVecFFE(g[i]),", \n");
    od;
    AppendTo(file,"]*One(GF(",String(p),")); \n\n");
    c := IntVecFFE(CoefficientsOfUnivariatePolynomial(L!.type));
    AppendTo(file,"LIEALG.type := ",c,"; \n\n");

    #######################################################################
    # add pres 
    #AddPresDB(p,d,nr);
	r := PresentationByGens(L, L!.gens);

    # write into file
    AppendTo(file, "LIEALG.rels := [\n");
    for i in [1..Length(r)] do
        AppendTo(file, r[i]);
        AppendTo(file, ",\n");
    od;
    AppendTo(file, "];\n\n");

    #######################################################################
    # that's it
    LIEALG := false;
end;

#############################################################################
CheckOutDB := function(p, dim, nr)
    local file, L, SLT, SLD, y, f;

    # get info
    if p = 2 then SLT := SLTB2; elif p = 3 then SLT := SLTB3; fi;
    if p = 2 then SLD := SLDB2; elif p = 3 then SLD := SLDB3; fi;

    # check
    if dim > Length(SLT) or dim < 0 or nr > SLT[dim] or nr < 0 then 
        Error("Lie Alg does not exists in DB"); 
    fi;

    # file name
    file := Concatenation(DBPATH,String(p),"/Dim_",String(dim),"_Nr_",
                          String(nr));

    # read
    Read(file);

    # create Lie Algebra
    L := LieAlgebraByTable(LIEALG.table);

    # add its known names
    L!.names := LIEALG.names;

    # add its invariants
    L!.nrnil := LIEALG.nrnil;
    L!.nrsol := LIEALG.nrsol;
    if IsBound(LIEALG.nrtyp) then L!.nrtyp := LIEALG.nrtyp; fi;
	if IsBound(LIEALG.res) then L!.res := LIEALG.res; fi;
	if IsBound(LIEALG.der) then L!.der := LIEALG.der; fi;
	if IsBound(LIEALG.ann) then L!.ann := LIEALG.ann; fi;
	if IsBound(LIEALG.exp) then L!.exp := LIEALG.exp; fi;
	if IsBound(LIEALG.aut) then L!.aut := LIEALG.aut; fi;

    # generators
    L!.gens := LIEALG.gens; 

    # type
    y := Indeterminate(GF(p), "y": old);
    f := List([0..Length(LIEALG.type)-1], x -> y^x);
    L!.type := LIEALG.type*f;

    # presentation
    L!.rels := LIEALG.rels; 

    # store and return
    SLD[dim][nr] := L;
    return L;
end;

ReadDB := function()
    local i, j;
    for i in [1..Length(SLTB2)] do
        if not IsBound(SLDB2[i]) then SLDB2[i] := []; fi;
        for j in [1..SLTB2[i]] do
            if not IsBound(SLDB2[i][j]) then
                CheckOutDB(2,i,j);
            fi;
        od;
    od;
    for i in [1..Length(SLTB3)] do
        if not IsBound(SLDB3[i]) then SLDB3[i] := []; fi;
        for j in [1..SLTB3[i]] do
            if not IsBound(SLDB3[i][j]) then
                CheckOutDB(2,i,j);
            fi;
        od;
    od;
end;

PrintDBByDim := function(p, i)
    local j, SLB;
    if p = 2 then SLB := SLDB2[i]; elif p = 3 then SLB := SLDB3[i]; fi;
    for j in [1..Length(SLB)] do
        Print("  ",j);
        if IsBound(SLB[j]!.nrnil) then 
            Print(", nrnil=",SLB[j]!.nrnil);
        fi;
        if IsBound(SLB[j]!.nrsol) then 
            Print(", nrsol=",SLB[j]!.nrsol);
        fi;
        if IsBound(SLB[j]!.aut) then 
            Print(", aut=",SLB[j]!.aut);
        fi;
        if Length(SLB[j]!.names) > 0 then 
            Print(" ",SLB[j]!.names);
        fi;
        Print("\n");
    od;
end;

PrintDB := function(p)
    local i, j, SLT;
    if p = 2 then SLT := SLTB2; elif p = 3 then SLT := SLTB3; fi;
    Print("\n");
    for i in [1..Length(SLT)] do
        if SLT[i] > 0 then 
            Print("Dimension ",i,":\n");
            PrintDBByDim(p, i);
        fi;
    od;
end;


